import os
from decimal import Decimal
from typing import List, Tuple, Dict, Union

from parsee.extraction.extractor_elements import ExtractedEl, StructuredTable, ExtractedSource, StructuredRow, \
    StructuredTableCell
from pdf_reader import get_elements_from_pdf
from parsee.converters.interfaces import RawToJsonConverter
from parsee.utils.enums import DocumentType, ElementType
from pdf_reader.custom_dataclasses import RelativeAreaPrediction
from pdf_reader.custom_dataclasses import ExtractedPage, ExtractedTable, ExtractedPdfElement, ExtractedFigure, LineItem
from parsee.utils.helper import is_year_cell, is_number_cell


PRICING_PDF_CONVERSION = Decimal(os.getenv("PRICING_CONVERSION")) if os.getenv("PRICING_CONVERSION") is not None else Decimal(0)


class PdfConverter(RawToJsonConverter):

    def __init__(self, predicted_areas: Union[None, Dict[int, List[RelativeAreaPrediction]]]):
        super().__init__(DocumentType.PDF)
        self.service_name = "parsee_pdf"
        self.areas = predicted_areas

    def pages_to_extracted_el(self, pages: List[ExtractedPage]) -> List[ExtractedEl]:
        elements: List[ExtractedEl] = []
        for page_num, page in enumerate(pages):
            for idx, el in enumerate(page.paragraphs):
                if isinstance(el, ExtractedTable):
                    element = self._make_structured_table(
                        ExtractedSource(DocumentType.PDF, {"x0": el.x0, "x1": el.x1, "y0": el.y0, "y1": el.y1}, None,
                                        len(elements), {"page_idx": page_num, "page_size": [page.size.x0, page.size.y0, page.size.x1, page.size.y1]}), el)
                elif isinstance(el, ExtractedFigure):
                    element = ExtractedEl(ElementType.FIGURE, ExtractedSource(DocumentType.PDF, {"x0": el.x0, "x1": el.x1, "y0": el.y0, "y1": el.y1}, None, len(elements),
                                                                              {"page_idx": page_num, "page_size": [page.size.x0, page.size.y0, page.size.x1, page.size.y1]}), None)
                elif isinstance(el, ExtractedPdfElement):
                    element = ExtractedEl(ElementType.TEXT, ExtractedSource(DocumentType.PDF,
                                                                            {"x0": el.x0, "x1": el.x1, "y0": el.y0,
                                                                             "y1": el.y1}, None, len(elements),
                                                                            {"page_idx": page_num,
                                                                             "page_size": [page.size.x0, page.size.y0, page.size.x1, page.size.y1]}), el.get_text())
                else:
                    raise Exception("element not recognized")
                elements.append(element)
        return elements

    def convert(self, file_path_or_content: str, **kwargs) -> Tuple[List[ExtractedEl], Decimal]:

        pages = get_elements_from_pdf(file_path_or_content, self.areas, **kwargs)

        return self.pages_to_extracted_el(pages), PRICING_PDF_CONVERSION

    def _make_structured_table(self, source: ExtractedSource, table: ExtractedTable) -> StructuredTable:

        rows_structured: List[StructuredRow] = []

        # determine header end
        header_end_idx = 0
        for k, line_item in enumerate(table.items):
            found_valid = False
            for v in line_item.values:
                if is_number_cell(v.val) and not is_year_cell(v.val):
                    found_valid = True
                    break
            if found_valid:
                header_end_idx = k
                break

        # add value rows
        for k, row in enumerate(table.items):
            rows_structured.append(self._make_structured_row("body" if k >= header_end_idx else "header", row))

        return StructuredTable(source, rows_structured)

    def _make_structured_row(self, row_type: str, row: LineItem) -> StructuredRow:
        row_values: List[StructuredTableCell] = []
        li_caption_cell = StructuredTableCell(row.caption)
        row_values.append(li_caption_cell)
        for v in row.values:
            val_obj = StructuredTableCell(v.val)
            if val_obj.valid:
                row_values.append(val_obj)
        return StructuredRow(row_type, row_values)



from parsee.datasets.dataset_dataclasses import DatasetRow


class Prompt:

    def __init__(self, description: str, main_task: str, additional_info: str, full_example: str, available_text: str):

        self.description = description
        self.main_task = main_task
        self.additional_info = additional_info
        self.full_example = full_example
        self.available_text = available_text

    def __str__(self) -> str:
        return f'''{self.description} \n

                    {self.main_task} \n
                    
                    {self.additional_info} \n

                    {self.full_example} \n
                    
                    This is the available data to fulfill the task:\n
                    {self.available_text}
                    '''




MERGE_MIN_CONFIDENCE = 0.6
PARTIAL_MIN_CONFIDENCE = 0.1
THRESHOLD_INBETWEEN_MERGE = 0.7
TEXT_DISTANCE_MERGE_THRESHOLD = 1000
TEXT_DISTANCE_CLOSE_STATEMENT_DETECTION = TEXT_DISTANCE_MERGE_THRESHOLD
MAX_DISTANCE_UNIQUE_PROXIMITY = 20000
ELEMENTS_WORDS_TO_INCLUDE = 50
ELEMENTS_TABLES_TO_INCLUDE = 3
NUM_TOKENS_DEFAULT_LLM = 4000

ID_NOT_AVAILABLE = "[NOT_SPECIFIED]"
ID_NOT_CONFIGURED = "[NOT_CONFIGURED]"



from pydantic.dataclasses import dataclass
from typing import Dict, List, Any, Union, Optional

from parsee.utils.enums import SearchStrategy
from parsee.templates.mappings import MappingSchema
from parsee.templates.general_structuring_schema import StructuringItemSchema


@dataclass
class ElementSchema:
    id: str
    title: str
    additionalInfo: str
    keywords: str
    takeBestInProximity: bool
    model: Optional[str]
    searchStrategy: SearchStrategy
    mapRows: Union[None, MappingSchema]
    mappingModel: Optional[str]
    metaInfoIds: List[str]

    def to_json_dict(self) -> Dict:
        return {"id": self.id, "title": self.title, "additionalInfo": self.additionalInfo, "keywords": self.keywords,
                "takeBestInProximity": self.takeBestInProximity, "model": self.model, "searchStrategy": self.searchStrategy.value,
                "mapRows": self.mapRows.to_json_dict() if self.mapRows is not None else None,
                "mappingModel": self.mappingModel,
                "metaInfoIds": self.metaInfoIds
                }


@dataclass
class ElementDetectionSchema:
    items: List[ElementSchema]
    settings: Dict[str, Any]

    def to_json_dict(self) -> Dict:
        return {"items": [x.to_json_dict() for x in self.items], "settings": self.settings}



import json
import tempfile
from decimal import Decimal
from typing import Tuple, Union, List, Dict

from parsee.utils.enums import DocumentType
from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.converters.json_to_raw import load_document_from_json
from parsee.converters.interfaces import RawToJsonConverter
from parsee.converters.html_extraction import HtmlConverter
from parsee.converters.pdf_extraction import PdfConverter
from parsee.converters.simple_text import SimpleTextConverter
from parsee.utils.helper import get_source_identifier, get_source_identifier_simple


def determine_document_type(file_path: str) -> DocumentType:
    file_path = file_path.lower()
    if file_path.endswith(".pdf") or file_path.endswith(".png") or file_path.endswith(".jpg") or file_path.endswith(".jpeg"):
        return DocumentType.PDF
    elif file_path.endswith(".html") or file_path.endswith(".xml"):
        return DocumentType.HTML
    else:
        raise Exception("document format not recognized")


def choose_converter(source_type: DocumentType) -> RawToJsonConverter:
    if source_type == DocumentType.HTML:
        return HtmlConverter()
    elif source_type == DocumentType.PDF:
        return PdfConverter(None)
    else:
        raise Exception("Unknown format")


def load_document(file_path: str) -> StandardDocumentFormat:
    doc_type = determine_document_type(file_path)
    source_identifier = get_source_identifier(file_path)
    doc, _ = doc_to_standard_format(source_identifier, doc_type, choose_converter(doc_type), file_path)
    return doc


def from_text(text: str) -> StandardDocumentFormat:
    source_identifier = get_source_identifier_simple(text)
    converter = SimpleTextConverter(DocumentType.TEXT)
    doc, _ = doc_to_standard_format(source_identifier, DocumentType.TEXT, converter, text)
    return doc


def doc_to_standard_format(source_identifier: str, source_type: DocumentType, converter: RawToJsonConverter,
                           file_path_or_content: str) -> Tuple[StandardDocumentFormat, Decimal]:
    elements, amount = converter.convert(file_path_or_content)
    return StandardDocumentFormat(source_type, source_identifier, elements), amount


def save_doc_in_standard_format(source_identifier: str, source_type: DocumentType, converter: RawToJsonConverter, file_path: str) -> \
        Tuple[any, StandardDocumentFormat, Decimal]:
    doc, amount = doc_to_standard_format(source_identifier, source_type, converter, file_path)
    json_string = json.dumps(doc.to_json_dict())
    tmp = tempfile.NamedTemporaryFile(delete=True)
    tmp.write(str.encode(json_string, 'utf-8'))
    return tmp, doc, amount


def load_standard_document_from_file(json_file_path: str) -> StandardDocumentFormat:
    with open(json_file_path, "r") as f:
        contents = json.loads(f.read())
        return load_document_from_json(contents)



import re
from typing import *

from parsee.utils.helper import clean_numeric_value, words_contained
from parsee.extraction.extractor_dataclasses import ParseeMeta, ExtractedSource


MAIN_QUESTION_STR = "(main question)"


def get_cleaned_list(values: List[str]):
    values_cleaned = []
    for x in values:
        if '-' in x:
            for y in x.split('-'):
                values_cleaned.append(clean_numeric_value(y))
        else:
            values_cleaned.append(clean_numeric_value(x))
    return values_cleaned


def parse_sources(prompt_answer: str, total_elements: Optional[int]) -> Tuple[str, List[int]]:
    result = re.search(r'((Source(s|):\s*|)(\[([\d-]+)\](,\s*|))+)', prompt_answer)
    sources = []
    final_answer = prompt_answer

    if result is not None:
        values_cleaned = get_cleaned_list(result.group().split("]"))
        sources = [int(x) for x in values_cleaned if x is not None and (total_elements is None or int(x) < total_elements)]
        final_answer = final_answer.replace(result.group(), "")
    else:
        # try with outer brackets only
        result = re.search(r'((Source(s|):\s*|)\[(\d(, |,|-|))+(\]|))', prompt_answer)
        if result is not None:
            values_cleaned = get_cleaned_list(result.group().split(","))
            sources = [int(x) for x in values_cleaned if x is not None and (total_elements is None or int(x) < total_elements)]
            final_answer = final_answer.replace(result.group(), "")
        else:
            # check if just 'Sources(s):' is present and remove if so
            result = re.search(r'(Source(s|):\s*)', prompt_answer)
            if result is not None:
                final_answer = final_answer.replace(result.group(), "")

    return final_answer, sources


def parse_answer_blocks(prompt_answer: str) -> List[str]:
    blocks = prompt_answer.split("\n\n")
    return [x for x in blocks if x.strip() != ""]


def parse_main_and_meta(prompt_answer_block: str) -> Dict[Union[None, str], str]:

    # mistral uses these characters sometimes
    prompt_answer_block = prompt_answer_block.replace("）", ") ")
    prompt_answer_block = prompt_answer_block.replace("（", " (")
    prompt_answer_block = prompt_answer_block.replace("：", ": ")

    # go line by line
    by_lines = prompt_answer_block.split("\n")
    output = {}
    for line in by_lines:
        match = re.search(r'\s*(\(.+\)\s*:|:|)([^\n]+)', line)
        if match is not None:
            meta_id = match.group(1) if len(words_contained(match.group(1))) > 0 else None
            value = match.group(2)
            if meta_id not in output:
                output[meta_id] = value
    return output


import re
from decimal import Decimal
from typing import Tuple, List, Union
import os

from bs4 import BeautifulSoup

from parsee.extraction.extractor_elements import ExtractedEl, StructuredTable
from parsee.converters.interfaces import RawToJsonConverter
from parsee.extraction.extractor_elements import ExtractedSource, StructuredRow, StructuredTableCell
from parsee.utils.enums import DocumentType, ElementType


PRICING_HTML_CONVERSION = Decimal(os.getenv("PRICING_CONVERSION")) if os.getenv("PRICING_CONVERSION") is not None else Decimal(0)


def get_element_sibling_xpath(element, xpath=""):
    parent = element.parent
    if parent is None:
        return xpath
    siblings = parent.find_all(element.name, recursive=False)
    if len(siblings) == 1:
        xpath = '/' + element.name + xpath
    else:
        index = siblings.index(element)
        xpath = f'/{element.name}[{index}]' + xpath
    return get_element_sibling_xpath(parent, xpath)


class HtmlConverter(RawToJsonConverter):

    def __init__(self):
        super().__init__(DocumentType.HTML)
        self.service_name = "parsee_html"

    def convert(self, file_path_or_content: str) -> Tuple[List[ExtractedEl], Decimal]:
        with open(file_path_or_content) as html_file:
            html_content = html_file.read()
            return self._extract_elements_from_html_data(html_content), PRICING_HTML_CONVERSION

    def _extract_elements_from_html_data(self, html_data) -> List[ExtractedEl]:

        soup = BeautifulSoup(html_data, 'lxml')

        # remove xbrli tags
        xbrli_tags = soup.find_all(re.compile(r'(xbrli:.+)'))
        for tag in xbrli_tags:
            tag.decompose()

        # remove line break tags and replace with space
        br_tags = soup.find_all('br')
        for tag in br_tags:
            tag.replace_with(" ")

        return self._extract_from_soup(soup)

    # extracts elements from beautifulsoup object
    def _extract_from_soup(self, soup, parent_list_len=0):
        # currently if there is text before a table NOT INSIDE SOME ELEMENT, this text is lost (this is considered minor)
        el_list = []

        # basic check if element is displayed or not
        style_str = soup["style"].replace(" ", "").lower() if "style" in soup.attrs else ""
        if "display:none" in style_str:
            return []

        # check if element contains a table
        check = soup.find_all("table")
        if len(check) > 0:
            # go down element by element
            # direct children
            direct_children = soup.find_all(recursive=False)
            for child in direct_children:
                if child.name == "table":
                    element_index = parent_list_len + len(el_list)
                    el_list.append(self._make_structured_table(
                        ExtractedSource(DocumentType.HTML, None, get_element_sibling_xpath(child), element_index, None), child))
                else:
                    # child is not a table, call function again
                    el_list += self._extract_from_soup(child, len(el_list) + parent_list_len)
            # lastly, add also text that is not contained in a child element
            text_pieces_remaining = [x.strip() for x in soup.find_all(text=True, recursive=False) if x.strip() != ""]
            base_xpath = get_element_sibling_xpath(soup)
            for text_piece in text_pieces_remaining:
                element_index = parent_list_len + len(el_list)
                el_list.append(
                    ExtractedEl(ElementType.TEXT, ExtractedSource(DocumentType.HTML, None, base_xpath, element_index, None),
                                text_piece))
        else:
            # no table inside, just return all contained strings
            element_index = parent_list_len + len(el_list)
            full_string_value = (" ".join(soup.stripped_strings)).strip()
            if full_string_value != "":
                el_list.append(
                    ExtractedEl(ElementType.TEXT, ExtractedSource(DocumentType.HTML, None, get_element_sibling_xpath(soup), element_index, None),
                                full_string_value))

        return el_list

    def _make_structured_table(self, source: ExtractedSource, soup) -> StructuredTable:

        # check for header
        header = soup.find("thead", recursive=False)
        rows_structured: List[StructuredRow] = []

        if header is not None:
            rows = header.find_all("tr")
            for row in rows:
                rows_structured.append(self._make_structured_row("header", row))

        # check for body
        body = soup.find("tbody", recursive=False)

        if body is not None:
            rows = body.find_all("tr")
            for row in rows:
                rows_structured.append(self._make_structured_row("body", row))

        # rows not inside thead or tbody
        rows = soup.find_all("tr", recursive=False)
        for row in rows:
            rows_structured.append(self._make_structured_row("body", row))

        # connect cells and determine amount of cols etc.
        return StructuredTable(source, rows_structured)

    def _make_structured_row(self, row_type: str, soup) -> StructuredRow:

        values = soup.find_all(recursive=False)
        structured_values: List[StructuredTableCell] = []

        for value in values:
            val_obj = self._make_structured_cell(value)
            if val_obj.valid:
                structured_values.append(val_obj)
        return StructuredRow(row_type, structured_values)

    def _make_structured_cell(self, soup) -> StructuredTableCell:

        val = soup.get_text()
        colspan = 1

        if "colspan" in soup.attrs and str(soup['colspan']).isnumeric():
            colspan = int(soup['colspan'])

        cell = StructuredTableCell(val, colspan)

        # check if object is valid
        style_str = soup["style"].replace(" ", "").lower() if "style" in soup.attrs else ""

        if "display:none" in style_str:
            cell.valid = False

        return cell



from decimal import Decimal
from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat, FinalOutputTableColumn, ParseeAnswer
from parsee.utils.enums import DocumentType
from parsee.templates.job_template import JobTemplate
from parsee.extraction.extractor_dataclasses import ParseeBucket
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.datasets.readers.interfaces import ModelReader
from parsee.storage.vector_stores.interfaces import VectorStore


class StorageManager:

    vector_store: VectorStore

    def __init__(self, vector_store: VectorStore):
        self.vector_store = vector_store

    def db_values_template(self, job_template: JobTemplate, strict: bool) -> JobTemplate:
        raise NotImplemented

    def log_expense(self, service: str, amount: Decimal, class_id: str):
        raise NotImplemented

    def get_available_models(self) -> List[MlModelSpecification]:
        raise NotImplemented
